use strict;
use warnings;

#application to locate where the perl source is so we can look in the same directory for other necessary files
use FindBin;
use lib $FindBin::Bin;

use Geo::ShapeFile;

#get a case insensitive glob whether we are on windows or unix
use File::Glob qw(:globally :nocase);

#get min and max functions for arrays
use List::Util qw(min max);

use rounding_functions;

my $map_dir = "binary_maps\\";


my $debug = 0;
my $long_straddle_mode = 0;

sub locate_shapefiles {
	my ($long,$lat) = @_;
	my $file_template = "binary_maps/*.shp";
	my @map_files = glob($file_template);
	my @shapefiles_to_return;
	foreach my $filename (@map_files) {
	   my ($shapefile_name) = $filename =~ /(.*)\.shp$/;
       push @shapefiles_to_return,$shapefile_name;
	}
	if (!scalar(@shapefiles_to_return)) {
	   die "No shapefiles found!!!";
	}else{
		return @shapefiles_to_return;
	}
}
my $debug_num = 0;
my $debugfile = "binary_maps/Europe_SWAsia";
sub output_coastlines {
	my ($filename, $shapefiles,$output_all_points,$min_long,$max_long,$min_lat,$max_lat) = @_; #set output to file to false to just see stats
	# note that longitude direction = x axis so output order is long,lat
	# also, lat and long are in degrees
	open (my $outfile, $filename) if $filename; #if filename is empty nothing will be output (use for debugging)
	open (my $debug_file, ">landmap_debug.csv") if $debug; #out here so visible below by 'strict'
	if ($debug) {
        print $outfile "longitude,latitude,file,shape,num_parts,point_num\n";
 	}else {
        print $outfile "longitude,latitude\n";
	}

	my @shapefiles = @$shapefiles;
	my $n_points_read = 0;
	my $n_shapes_read = 0;
	my $n_points_output = 0;
	my $outputting_points = 0;
	foreach my $shapefile_name (@shapefiles) {
		my $shapefile = new Geo::ShapeFile($shapefile_name);
		print "processing $shapefile_name\n";
		if ($debug and $shapefile_name eq $debugfile) {
			print "In map to debug\n";
		}
		my $output_started = 0;
		#print "$total_shapes shapes located to process\n";
		my @shapes_to_test;
		if ($output_all_points) {
           	@shapes_to_test = (1..$shapefile->shapes);
		}else{
			if ($long_straddle_mode) {
				@shapes_to_test = ($shapefile->shapes_in_area($min_long,$min_lat,180,$max_lat),
					$shapefile->shapes_in_area(-180,$min_lat,$max_long,$max_lat));
			} else {
                @shapes_to_test = ($shapefile->shapes_in_area($min_long,$min_lat,$max_long,$max_lat));
			}
		}
		foreach my $shapeid (@shapes_to_test) {

			if ($debug and $debug_num == $shapeid and $shapefile_name eq $debugfile) {
				print "At debug shape\n"; #useful for debugging
			}
			my $shape = $shapefile->get_shp_record($shapeid);
			#if (!$output_all_points) {
			#	next if ($max_long < $shape->x_min or $min_long > $shape->x_max or $max_lat < $shape->y_min or $min_lat > $shape->y_max);
			#}
			#if ($output_started) {
			#	print "$filename has some points in the required area\n";
			#}

			my $num_pts_in_shape = 0; #for debugging

			if ($filename) {
				foreach my $part (1..$shape->num_parts) {
					my $part_ptr = [($shape->get_part($part))];
					foreach my $point (@$part_ptr) {
						$num_pts_in_shape++;
						my ( $point_long,$point_lat ) = ( $point->X,$point->Y );
			            if ($output_all_points

							or ( !$long_straddle_mode and
								($min_long <= $point_long and $point_long <= $max_long and $min_lat <= $point_lat and $point_lat <= $max_lat)
							   )
							or ( $long_straddle_mode and
								 	(	($min_long <= $point_long and $point_long <= 180 and $min_lat <= $point_lat and $point_lat <= $max_lat)
								 		or
								 	    (-180 <= $point_long and $point_long <= $max_long and $min_lat <= $point_lat and $point_lat <= $max_lat)
								 	)
							   )
							)
							{
							if (!$outputting_points) {
			                    print $outfile "0,\n" unless !$n_points_output; #no blank line before very first output position
								#a zero and a blank handily makes a gap in an x/y line chart when all the points are plotted as one series in excel
								$outputting_points = 1;
							}
			                if ($debug) {
	                            print $outfile join(',',$point->X,$point->Y,$shapefile_name,$shapeid,$shape->num_parts,$num_pts_in_shape)."\n";
							}else{
								print $outfile $point->X.','.$point->Y."\n";
							}
			                $n_points_output++;
			            } else {
							$outputting_points = 0;
						}
					}
					#blank needed if we output any more points
					$outputting_points = 0;
				}
				print $debug_file join(',',$shapefile->{filebase},$shapeid,$num_pts_in_shape)."\n" if $debug;

			}
			$n_points_read += $shape->num_points();
			$n_shapes_read ++;

		}
	}
	print "Processed $n_shapes_read coastline parts, considered $n_points_read points and $n_points_output of them were output\n,";
}

my ($long_min,$lat_min,$long_max,$lat_max);

print "Input min and max lat and long in degrees, space delimited, eg \"48 50 -2 2\", return to exit\n";
print "Or a single lat and long in degrees to set the centre\n";
chomp(my $min_max_lat_long = <>);
if ($min_max_lat_long) {
	($lat_min,$lat_max, $long_min, $long_max) = split(" ",$min_max_lat_long);
	if (!$long_max) {
		print "input the lat and long width of the map required, in degrees, eg \"4 3\"\n";
		chomp(my $widths = <>);
		my ($lat_width, $long_width) = split(" ",$widths);
       ($lat_min,$lat_max, $long_min, $long_max) =
	   		($lat_min - $lat_width/2, $lat_min + $lat_width/2, $long_min - $long_width/2, $long_min + $long_width/2);
	}
	if ($long_min > $long_max) {
		print "Map straddles 180 degrees longitude, special mode entered\n";
		$long_straddle_mode = 1;
	}
}else{
	print "no lat long given, exiting\n";
	exit;
}

print "Input name for this map, e.g \"iow\"\n";
chomp(my $map_name = <>);

$map_name = ">".$map_name."_".join("_",round($lat_min),round($lat_max),round($long_min),round($long_max)).".csv";

my @shapefiles = locate_shapefiles(($long_min + $long_max)/2,($lat_min + $lat_max)/2);

output_coastlines($map_name, \@shapefiles,0,$long_min,$long_max,$lat_min,$lat_max);
